// Copyright 2019, University of Freiburg,
// Chair of Algorithms and Data Structures.
// Authors: Patrick Brosi <brosi@informatik.uni-freiburg.de>

#include <fstream>
#include <iostream>
#include <set>
#include <sstream>
#include <string>
#include "osmfixer/index/StatIdx.h"
#include "util/geo/BezierCurve.h"

using osmfixer::StatIdx;
using osmfixer::OsmAttrs;
using osmfixer::Station;
using osmfixer::Group;
using osmfixer::Suggestion;

// _____________________________________________________________________________
void StatIdx::readFromFile(const std::string& path) {
  std::ifstream f(path);

  std::string line;

  // first, parse nodes
  while (std::getline(f, line)) {
    // empty line is separator between blocks
    if (line.size() == 0) break;

    std::stringstream rec(line);
    size_t osmid, origGroup, group;
    double lat, lng;
    OsmAttrs attrs;

    rec >> osmid;
    rec >> lat;
    rec >> lng;
    rec >> origGroup;
    rec >> group;

    std::string temp;
    std::getline(rec, temp, '\t');
    while (std::getline(rec, temp, '\t')) {
      std::string key, val;
      key = temp;
      std::getline(rec, temp, '\t');
      val = temp;
      attrs[key].push_back(val);
    }

    addStation(osmid, lat, lng, origGroup, group, attrs);
  }

  // second, parse groups
  while (std::getline(f, line)) {
    // empty line is separator between blocks
    if (line.size() == 0) break;

    std::stringstream rec(line);
    size_t osmid;
    OsmAttrs attrs;

    rec >> osmid;

    addGroup(osmid, attrs);
  }

  // third, parse attr errs
  while (std::getline(f, line)) {
    // empty line is separator between blocks
    if (line.size() == 0) break;

    std::stringstream rec(line);
    size_t id;
    std::string attr, otherAttr;
    double conf;
    size_t otherId;

    rec >> id;

    std::getline(rec, attr, '\t');
    std::getline(rec, attr, '\t');
    std::getline(rec, otherAttr, '\t');

    rec >> conf;
    rec >> otherId;

    _stations[id].attrErrs.push_back({attr, otherAttr, conf, otherId});
  }

  initGroups();
  initSuggestions();
  initIndex();
}

// _____________________________________________________________________________
void StatIdx::addStation(size_t osmid, double lat, double lng, size_t origGroup,
                         size_t group, const OsmAttrs& attrs) {
  auto point = util::geo::latLngToWebMerc<double>(lat, lng);

  _stations.emplace_back(
      Station{_stations.size(), osmid, origGroup, group, point, attrs});

  // extend bounding box
  _bbox = util::geo::extendBox(point, _bbox);
}

// _____________________________________________________________________________
void StatIdx::addGroup(size_t osmid, const OsmAttrs& attrs) {
  Group g;
  g.id = _groups.size();
  g.osmid = osmid;
  g.attrs = attrs;
  _groups.emplace_back(g);
}

// _____________________________________________________________________________
void StatIdx::initGroups() {
  for (size_t i = 0; i < _stations.size(); i++) {
    // this should be ensured by the input file
    assert(_stations[i].origGroup < _groups.size());
    _groups[_stations[i].origGroup].stations.push_back(i);

    assert(_stations[i].group < _groups.size());
    if (_stations[i].group != _stations[i].origGroup &&
        _groups[_stations[i].group].osmid == 1) {
      // only add non-orig groups to polygons of new (non-osm) groups
      _groups[_stations[i].group].stations.push_back(i);
    }
  }

  for (size_t i = 0; i < _groups.size(); i++) {
    util::geo::MultiPoint<double> mp;
    for (size_t stid : _groups[i].stations) {
      double rad = 11.0;
      int n = 20;
      for (int i = 0; i < n; i++) {
        double x = rad * cos((2.0 * M_PI / static_cast<double>(n)) *
                             static_cast<double>(i));
        double y = rad * sin((2.0 * M_PI / static_cast<double>(n)) *
                             static_cast<double>(i));

        mp.push_back(util::geo::DPoint(_stations[stid].pos.getX() + x,
                                       _stations[stid].pos.getY() + y));
      }
    }
    _groups[i].poly = util::geo::convexHull(mp);
  }
}

// _____________________________________________________________________________
void StatIdx::initIndex() {
  _sgrid = util::geo::Grid<size_t, util::geo::Point, double>(5000, 5000, _bbox,
                                                             false);
  _ggrid = util::geo::Grid<size_t, util::geo::Polygon, double>(5000, 5000,
                                                               _bbox, false);
  _suggrid = util::geo::Grid<size_t, util::geo::Line, double>(5000, 5000, _bbox,
                                                              false);
  for (size_t i = 0; i < _stations.size(); i++) _sgrid.add(_stations[i].pos, i);
  for (size_t i = 0; i < _groups.size(); i++) _ggrid.add(_groups[i].poly, i);
  for (size_t i = 0; i < _suggestions.size(); i++)
    _suggrid.add(_suggestions[i].arrow, i);
}

// _____________________________________________________________________________
std::vector<const Suggestion*> StatIdx::getSuggestions(
    const util::geo::DBox bbox) const {
  std::vector<const Suggestion*> ret;
  auto ll = util::geo::latLngToWebMerc<double>(bbox.getLowerLeft().getX(),
                                               bbox.getLowerLeft().getY());
  auto ur = util::geo::latLngToWebMerc<double>(bbox.getUpperRight().getX(),
                                               bbox.getUpperRight().getY());

  std::set<size_t> tmp;
  auto reqBox = util::geo::DBox(ll, ur);
  _suggrid.get(reqBox, &tmp);

  for (auto i : tmp) {
    if (util::geo::intersects(_suggestions[i].arrow, reqBox))
      ret.push_back(&_suggestions[i]);
  }

  return ret;
}

// _____________________________________________________________________________
std::vector<const Group*> StatIdx::getGroups(const util::geo::DBox bbox) const {
  std::vector<const Group*> ret;
  auto ll = util::geo::latLngToWebMerc<double>(bbox.getLowerLeft().getX(),
                                               bbox.getLowerLeft().getY());
  auto ur = util::geo::latLngToWebMerc<double>(bbox.getUpperRight().getX(),
                                               bbox.getUpperRight().getY());

  std::set<size_t> tmp;
  auto reqBox = util::geo::DBox(ll, ur);
  _ggrid.get(reqBox, &tmp);

  for (auto i : tmp) {
    if (util::geo::intersects(_groups[i].poly, reqBox))
      ret.push_back(&_groups[i]);
  }

  return ret;
}

// _____________________________________________________________________________
std::vector<const Station*> StatIdx::getStations(
    const util::geo::DBox bbox) const {
  std::vector<const Station*> ret;
  auto ll = util::geo::latLngToWebMerc<double>(bbox.getLowerLeft().getX(),
                                               bbox.getLowerLeft().getY());
  auto ur = util::geo::latLngToWebMerc<double>(bbox.getUpperRight().getX(),
                                               bbox.getUpperRight().getY());

  std::set<size_t> tmp;
  auto reqBox = util::geo::DBox(ll, ur);
  _sgrid.get(reqBox, &tmp);

  for (auto i : tmp) {
    if (util::geo::contains(_stations[i].pos, reqBox))
      ret.push_back(&_stations[i]);
  }

  return ret;
}

// _____________________________________________________________________________
const Station* StatIdx::getStation(size_t id) const {
  if (id >= _stations.size()) return 0;
  return &_stations[id];
}

// _____________________________________________________________________________
void StatIdx::initSuggestions() {
  for (size_t i = 0; i < _stations.size(); i++) {
    auto& stat = _stations[i];
    if (stat.group != stat.origGroup || getGroup(stat.group)->osmid == 1) {
      Suggestion sug;
      sug.station = i;
      sug.arrow = util::geo::DLine{stat.pos, stat.pos};
      if (getGroup(stat.origGroup)->osmid < 2) {
        if (getGroup(stat.group)->osmid == 1) {
          // move orphan into new group
          sug.type = 1;
          sug.target_gid = stat.group;

          // sug.arrow = getGroupArrow(i, stat.group);

          _suggestions.push_back(sug);
          stat.suggestions.push_back(_suggestions.size() - 1);

        } else if (getGroup(stat.group)->osmid > 1) {
          // move orphan into existing group
          sug.type = 2;
          sug.target_gid = stat.group;
          sug.target_osm_rel_id = getGroup(stat.group)->osmid;

          sug.arrow = getGroupArrow(i, stat.group);

          _suggestions.push_back(sug);
          stat.suggestions.push_back(_suggestions.size() - 1);
        }
      } else {
        if (getGroup(stat.group)->osmid == 1) {
          // move station from relation into new group
          sug.type = 3;
          sug.orig_gid = stat.origGroup;
          sug.orig_osm_rel_id = getGroup(stat.origGroup)->osmid;
          sug.target_gid = stat.group;

          sug.arrow = getGroupArrow(i, stat.group);

          _suggestions.push_back(sug);
          stat.suggestions.push_back(_suggestions.size() - 1);
        } else if (getGroup(stat.group)->osmid > 1) {
          // move station from relation into existing group
          sug.type = 4;
          sug.orig_gid = stat.origGroup;
          sug.orig_osm_rel_id = getGroup(stat.origGroup)->osmid;
          sug.target_gid = stat.group;
          sug.target_osm_rel_id = getGroup(stat.group)->osmid;

          _suggestions.push_back(sug);
          stat.suggestions.push_back(_suggestions.size() - 1);
        } else {
          // move station out of relation
          sug.type = 5;
          sug.orig_gid = stat.origGroup;
          sug.target_gid = stat.group;
          sug.target_osm_rel_id = getGroup(stat.group)->osmid;

          _suggestions.push_back(sug);
          stat.suggestions.push_back(_suggestions.size() - 1);
        }
      }
    }
  }
}

// _____________________________________________________________________________
util::geo::DLine StatIdx::getGroupArrow(size_t stat, size_t group) const {
  auto a = getStation(stat)->pos;
  auto b = util::geo::centroid(getGroup(group)->poly);

  auto pl = util::geo::PolyLine<double>(a, b);
  auto bb = pl.getPointAtDist(fmax(pl.getLength() / 2, pl.getLength() - 20)).p;

  if (util::geo::dist(a, bb) < 0.0001) {
    bb = a;
    bb.setX(bb.getX() + 10);
  }

  int side = rand() % 2;
  if (!side) side = -1;

  auto rot1 = util::geo::rotate(util::geo::DLine{a, bb}, 45 * side, a);
  auto rot2 = util::geo::rotate(util::geo::DLine{a, bb}, -20 * side, bb);

  auto rot3 = util::geo::rotate(util::geo::DLine{a, bb}, -120 * side, bb);

  auto i =
      util::geo::intersection(util::geo::LineSegment<double>{rot1[0], rot1[1]},
                              util::geo::LineSegment<double>{rot2[0], rot2[1]});

  auto line =
      util::geo::BezierCurve<double>(a, i, rot3[0], bb).render(1).getLine();

  // arrowhead
  auto pl2 = util::geo::PolyLine<double>(line);
  util::geo::LineSegment<double> headSeg(
      pl2.getPointAtDist(pl2.getLength() - 5).p, line.back());
  auto arrHeadLeft = util::geo::rotate(headSeg, -45, line.back()).first;
  auto arrHeadRight = util::geo::rotate(headSeg, 45, line.back()).first;

  auto tmp = line.back();

  line.push_back(arrHeadLeft);
  line.push_back(tmp);
  line.push_back(arrHeadRight);

  return line;
}

// _____________________________________________________________________________
const Group* StatIdx::getGroup(size_t id) const {
  if (id >= _groups.size()) return 0;
  return &_groups[id];
}

// _____________________________________________________________________________
const Suggestion* StatIdx::getSuggestion(size_t id) const {
  if (id >= _suggestions.size()) return 0;
  return &_suggestions[id];
}
